;=============================================================================
;                    Amiga 4-voice software mixer, by 8bitbubsy
;
; Features:
; - 32-bit fractional sampling precision
; - Works like Paula in some ways (len/dat regs updated after current cycle)
;
; Disadvantages:
; - No interpolation, but we don't want that in a Paula mixer anyway!
; - Mixing frequency must be above 31388Hz (single-step mixer, min. period=113)
;
; TABSIZE = 8 spaces
;=============================================================================

.386p
.MODEL FLAT

	INCLUDE "apaula.i"

.CODE

SMP_LEN_THRESHOLD EQU 32

	PUBLIC _Mixer_MixVoices
	PUBLIC _Mixer_GenerateVolTable

	; paula.c
	EXTERN _paulaOffset:DWORD

	; Input:
	;  EDI = Pointer to volume table (16-bit data)
	ALIGN 4
_Mixer_GenerateVolTable PROC NEAR
	push eax
	push ecx
	push edi
	std			; write in reverse
	; --------------------
	mov ecx,((PAULA_MAX_VOL+1)*256)-1
	lea edi,[edi+ecx*2]	; start at end of LUT
	; --------------------
	ALIGN 4
MGLoop:	mov eax,ecx
	imul ah			; ax = al(0..64, volume) * ah(-128..127, smpdata)
	stosw
	dec ecx
	jns MGLoop
	; --------------------
	cld
	pop edi
	pop ecx
	pop eax
	retn
_Mixer_GenerateVolTable ENDP

; ---------------------------------------------------------
; Description:
;  Regular mixing macros (optimized for i386DX)
;
; Register map:
;  EAX = <free>
;  EBX = Volume LUT @ current volume (lo-byte = smp, *2 = real address)
;  ECX = 32-bit sampling address delta (lo-byte = 16, for shifts)
;  EDX = 32-bit sampling address integer
;  ESI = Address to current voice struct
;  EDI = 16-bit mono mix buffer
;  EBP = 32-bit sampling address fraction
;  ESP = Zero (do not modify!)
; ---------------------------------------------------------

; used while unaligned
MIX1_M MACRO dstoffs
	mov bl,[edx]
	add ebp,ecx
	mov ax,[ebx+ebx]
	adc edx,esp
	add [edi+(dstoffs*2)],ax
ENDM

; used after aligned to a multiple of 32
MIX2_M MACRO dstoffs
	mov bl,[edx]
	add ebp,ecx
	mov ax,[ebx+ebx]
	adc edx,esp
	shl eax,cl
	; ----------------
	mov bl,[edx]
	add ebp,ecx
	mov ax,[ebx+ebx]
	adc edx,esp
	rol eax,cl ; word order fix :(
	; ----------------
	add [edi+(dstoffs*4)],eax
ENDM

; ---------------------------------------------------------
; Description:
;  Alternative mixing macros (optimized for i386).
;  Used when the sample length is short, to avoid overhead from
;  an end-of-sample calculation loop.
;
; Register map:
;  EAX = <free>
;  EBX = Volume LUT @ current volume (lo-byte = smp, *2 = real address)
;  ECX = 32-bit sampling address delta (lo-byte = 16, for shifts)
;  EDX = 32-bit sampling address integer
;  ESI = Address to current voice struct
;  EDI = 16-bit mono mix buffer
;  EBP = 32-bit sampling address fraction
;  ESP = Pointer to end of sample
; ---------------------------------------------------------

; alternative mixers for when sample length is <32 (less overhead)
AMIX_M MACRO dstoffs,jmpaddr
	mov bl,[edx]
	add ebp,ecx
	mov ax,[ebx+ebx]
	adc edx,0
	add [edi+(dstoffs*2)],ax
	cmp edx,esp 			; end of sample reached?
	jae jmpaddr
ENDM

EOS_M MACRO jmpaddr,endaddr
	mov esp,v_newLength[esi]
	cmp esp,2			; if new length is 2, stop mixing this voice
	jbe endaddr
	mov v_length[esi],esp
	mov edx,v_newDataPtr[esi]	; set new sample address
	add esp,edx			; set new sample end address
	jmp jmpaddr
ENDM
; ----------------------------------------------------------------

	; Input:
	;  EDI = 16-bit mono buffer
	;   CX = Number of samples to mix
	;
	; Output:
	;  AL = 1 if we mixed a voice, 0 if none were mixed
	ALIGN 4
_Mixer_MixVoices PROC NEAR
	pushad
	; we can safely take over ESP since the mixer is called from an
	; interrupt where no other interrupts can be fired.
	mov OldStackPointer,esp	
	; -------------------------
	mov byte ptr DidMixAVoice,0
	; -------------------------
	sub edi,32*2			; pre-sub buffer (for unroll alignment)	
	mov MixBufPtr,edi
	; -------------------------
	mov esi,_paulaOffset
	mov NumSamples,cx
	mov word ptr VoiceCounter,PAULA_VOICES-1
VoiceLoop:
	; -------------------------
	cmp byte ptr v_dmaActive[esi],0
	je InactiveVoice
	; -------------------------
	cmp dword ptr v_delta[esi],0
	je InactiveVoice
	; -------------------------
	; if looping 2-byte loop, we're inactive
	mov ebx,v_length[esi]
	cmp ebx,2
	ja Active
	cmp dword ptr v_newLength[esi],2
	jbe InactiveVoice
	; -------------------------
Active:	mov ax,NumSamples
	mov SamplesLeft,ax
	xor eax,eax
	mov MixBufIndex,eax
	mov byte ptr DidMixAVoice,1
	; -------------------------
	; If length is small, use alternative mixing
	; with less calculation overhead.
	; -------------------------
	cmp ebx,SMP_LEN_THRESHOLD
	jb AlternativeMix
	; -------------------------
	ALIGN 4
MixLoop:
	; ---------------------
	; Calculate max number of samples to mix
	; before reaching end of sample.
	; ---------------------
	mov edx,v_dataEnd[esi]
	sub edx,v_dataPtr[esi]
	dec edx
	mov eax,v_frac[esi]
	not eax
	div dword ptr v_delta[esi]
	inc eax
	mov cl,1
	movzx ebx,word ptr SamplesLeft
	cmp eax,ebx
	jle L1
	mov eax,ebx
	xor cl,cl
L1:	sub ebx,eax
	mov SamplesLeft,bx
	mov SmpEndReached,cl
	; ---------------------
	; Setup mix buffer pointer
	; ---------------------
	mov ecx,eax		; eax = SamplesToMix
	and ecx,32-1		; ecx = SamplesToMix % 32 (remainder)
	; ---------------------
	mov edi,MixBufPtr
	lea edi,[edi+ecx*2]	; pre-adjust output buffer for unrolled loop
	add edi,MixBufIndex
	lea ebx,[eax+eax]
	add MixBufIndex,ebx
	; ---------------------
	mov ebx,eax
	shr ebx,5		; SamplesToMix / 32 (for loop counter)
	mov SamplesToMix32,bx
	mov eax,[MixTable+ecx*4]
	; ---------------------
	xor esp,esp		; must be zero for adc instruction
	mov ebx,v_volTabPtr[esi]
	mov ecx,v_delta[esi]

	; Trash lo-byte of 32-bit delta so that CL can be used for bitshifts.
	; This saves two bytes of opcode data per iteration in the mixing loop.
	; The voice frequency will be off by up to 0.0025Hz from this hack
	; (i.e. you won't hear a difference).
	mov cl,16

	mov edx,v_dataPtr[esi]
	mov ebp,v_frac[esi]
	jmp eax ; do mixing (we end up at MixDone when done)
	
	ALIGN 4
AlternativeMix:
	movzx eax,word ptr SamplesLeft
	test eax,eax
	jz NextVoice
	; ---------------------
	; Setup mix buffer pointer
	; ---------------------
	mov ecx,eax		; eax = SamplesToMix
	and ecx,32-1		; ecx = SamplesToMix % 32 (remainder)
	; ---------------------
	mov edi,MixBufPtr
	lea edi,[edi+ecx*2]	; pre-adjust output buffer for unrolled loop
	add edi,MixBufIndex	
	lea ebx,[eax+eax]
	add MixBufIndex,ebx	
	; ---------------------
	mov ebx,eax
	shr ebx,5		; SamplesToMix / 32 (for loop counter)
	mov SamplesToMix32,bx
	mov eax,[AltMixTable+ecx*4]
	; ---------------------
	mov ebx,v_volTabPtr[esi]
	mov ecx,v_delta[esi]

	; hack that's not needed here, but keep it for hz error consistency
	mov cl,16	
	
	mov edx,v_dataPtr[esi]
	mov ebp,v_frac[esi]
	mov esp,v_dataEnd[esi]
	
	jmp eax ; do mixing
	
	ALIGN 4
MixTable:
	dd mix00,mix01,mix02,mix03,mix04,mix05,mix06,mix07
	dd mix08,mix09,mix0A,mix0B,mix0C,mix0D,mix0E,mix0F
	dd mix10,mix11,mix12,mix13,mix14,mix15,mix16,mix17
	dd mix18,mix19,mix1A,mix1B,mix1C,mix1D,mix1E,mix1F
	
AltMixTable:
	dd amix00,amix01,amix02,amix03,amix04,amix05,amix06,amix07
	dd amix08,amix09,amix0A,amix0B,amix0C,amix0D,amix0E,amix0F
	dd amix10,amix11,amix12,amix13,amix14,amix15,amix16,amix17
	dd amix18,amix19,amix1A,amix1B,amix1C,amix1D,amix1E,amix1F

; --------------------------------
; "End of sample" chunks
; --------------------------------
	ALIGN 4
eos00:	EOS_M amix00,EndSample
eos01:	EOS_M amix01,EndSample
eos02:	EOS_M amix02,EndSample
eos03:	EOS_M amix03,EndSample
eos04:	EOS_M amix04,EndSample
eos05:	EOS_M amix05,EndSample
eos06:	EOS_M amix06,EndSample
eos07:	EOS_M amix07,EndSample
eos08:	EOS_M amix08,EndSample
eos09:	EOS_M amix09,EndSample
eos0A:	EOS_M amix0A,EndSample
eos0B:	EOS_M amix0B,EndSample
eos0C:	EOS_M amix0C,EndSample
eos0D:	EOS_M amix0D,EndSample
eos0E:	EOS_M amix0E,EndSample
eos0F:	EOS_M amix0F,EndSample
eos10:	EOS_M amix10,EndSample
eos11:	EOS_M amix11,EndSample
eos12:	EOS_M amix12,EndSample
eos13:	EOS_M amix13,EndSample
eos14:	EOS_M amix14,EndSample
eos15:	EOS_M amix15,EndSample
eos16:	EOS_M amix16,EndSample
eos17:	EOS_M amix17,EndSample
eos18:	EOS_M amix18,EndSample
eos19:	EOS_M amix19,EndSample
eos1A:	EOS_M amix1A,EndSample
eos1B:	EOS_M amix1B,EndSample
eos1C:	EOS_M amix1C,EndSample
eos1D:	EOS_M amix1D,EndSample
eos1E:	EOS_M amix1E,EndSample
eos1F:	EOS_M amix1F,EndSample

; --------------------------------
; Mixing chunks
; --------------------------------
	ALIGN 4
amixL:	AMIX_M 00,eos1F
amix1F:	AMIX_M 01,eos1E
amix1E:	AMIX_M 02,eos1D
amix1D:	AMIX_M 03,eos1C
amix1C:	AMIX_M 04,eos1B
amix1B:	AMIX_M 05,eos1A
amix1A:	AMIX_M 06,eos19
amix19:	AMIX_M 07,eos18
amix18:	AMIX_M 08,eos17
amix17:	AMIX_M 09,eos16
amix16:	AMIX_M 10,eos15
amix15:	AMIX_M 11,eos14
amix14:	AMIX_M 12,eos13
amix13:	AMIX_M 13,eos12
amix12:	AMIX_M 14,eos11
amix11:	AMIX_M 15,eos10
amix10:	AMIX_M 16,eos0F
amix0F:	AMIX_M 17,eos0E
amix0E:	AMIX_M 18,eos0D
amix0D:	AMIX_M 19,eos0C
amix0C:	AMIX_M 20,eos0B
amix0B:	AMIX_M 21,eos0A
amix0A:	AMIX_M 22,eos09
amix09:	AMIX_M 23,eos08
amix08:	AMIX_M 24,eos07
amix07:	AMIX_M 25,eos06
amix06:	AMIX_M 26,eos05
amix05:	AMIX_M 27,eos04
amix04:	AMIX_M 28,eos03
amix03:	AMIX_M 29,eos02
amix02:	AMIX_M 30,eos01
amix01:	AMIX_M 31,eos00
amix00:	add edi,32*2
	dec word ptr SamplesToMix32
	jns amixL
	mov v_dataPtr[esi],edx
	mov v_frac[esi],ebp	
	mov v_dataEnd[esi],esp
	jmp NextVoice

	; --------------------------------------------------
	; Mix samples until SamplesToMix is a multiple of 32
	; --------------------------------------------------
	ALIGN 4
mix1F:	MIX1_M 01
mix1E:	MIX1_M 02
mix1D:	MIX1_M 03
mix1C:	MIX1_M 04
mix1B:	MIX1_M 05
mix1A:	MIX1_M 06
mix19:	MIX1_M 07
mix18:	MIX1_M 08
mix17:	MIX1_M 09
mix16:	MIX1_M 10
mix15:	MIX1_M 11
mix14:	MIX1_M 12
mix13:	MIX1_M 13
mix12:	MIX1_M 14
mix11:	MIX1_M 15
mix10:	MIX1_M 16
mix0F:	MIX1_M 17
mix0E:	MIX1_M 18
mix0D:	MIX1_M 19
mix0C:	MIX1_M 20
mix0B:	MIX1_M 21
mix0A:	MIX1_M 22
mix09:	MIX1_M 23
mix08:	MIX1_M 24
mix07:	MIX1_M 25
mix06:	MIX1_M 26
mix05:	MIX1_M 27
mix04:	MIX1_M 28
mix03:	MIX1_M 29
mix02:	MIX1_M 30
mix01:	MIX1_M 31
mix00:	add edi,32*2
	; ---------------------------------------------------------
	; SamplesToMix is now a multiple of 32, enter faster mixing
	; ---------------------------------------------------------
	cmp word ptr SamplesToMix32,0
	je MixDone
	ALIGN 4
InnerMixLoop:
	MIX2_M 00
	MIX2_M 01
	MIX2_M 02
	MIX2_M 03
	MIX2_M 04
	MIX2_M 05
	MIX2_M 06
	MIX2_M 07
	MIX2_M 08
	MIX2_M 09
	MIX2_M 10
	MIX2_M 11
	MIX2_M 12
	MIX2_M 13
	MIX2_M 14
	MIX2_M 15
	add edi,16*4
	dec word ptr SamplesToMix32
	jnz InnerMixLoop

MixDone:
	mov v_frac[esi],ebp	
	mov v_dataPtr[esi],edx
	; --------------------------------------------
	; End of sample reached?
	; --------------------------------------------
	cmp byte ptr SmpEndReached,0
	je NoEnd
	; --------------------------------------------
	; Yes, update Paula regs
	; --------------------------------------------
	mov eax,v_newLength[esi]
	mov v_length[esi],eax
	mov ebx,v_newDataPtr[esi]
	mov v_dataPtr[esi],ebx
	add ebx,eax
	mov v_dataEnd[esi],ebx
	; --------------------------------------------
	; If the sample has no loop, stop voice
	; --------------------------------------------
	cmp eax,2	; new length <= 2?
	ja NoEnd	; nope, this is a looping sample
EndSample:
	mov byte ptr v_dmaActive[esi],0
	jmp NextVoice
NoEnd:	; --------------------------------------------
	cmp eax,SMP_LEN_THRESHOLD
	jb AlternativeMix
	; --------------------------------------------
	cmp word ptr SamplesLeft,0
	jne MixLoop
	
NextVoice:
	add esi,v_sizeof
	dec word ptr VoiceCounter
	jns VoiceLoop
; ---------------------------------------------------------
	mov esp,OldStackPointer
	popad
	mov al,DidMixAVoice
	retn

InactiveVoice:
	mov ebx,v_newLength[esi]
	mov v_length[esi],ebx
	mov ecx,v_newDataPtr[esi]
	mov v_dataPtr[esi],ecx
	add ecx,ebx
	mov v_dataEnd[esi],ecx
	xor eax,eax
	mov v_frac[esi],eax
	jmp NextVoice

_Mixer_MixVoices ENDP

	ALIGN 4
OldStackPointer	dd 0
MixBufPtr	dd 0
MixBufIndex	dd 0
NumSamples	dw 0
SamplesLeft	dw 0
VoiceCounter	dw 0
SamplesToMix32	dw 0
SmpEndReached	db 0
DidMixAVoice	db 0

END
