// partly based on DMAW32.C from the official SB SDK

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <dos.h>
#include <conio.h>
#include "dma.h"

#define DMA_CHANNELS 8

typedef struct
{
	uint16_t length;
	void *buffer;
	uint8_t mode;
} dma_t;

static uint16_t bufferSelector;
static bool initialized[DMA_CHANNELS], active[DMA_CHANNELS];

// globals
const dmaports_t dmaPorts[8] =
{
	{ 0x00, 0x01, 0x87, 0x0A, 0x0B, 0x0C, 0x09, 0x0E },
	{ 0x02, 0x03, 0x83, 0x0A, 0x0B, 0x0C, 0x09, 0x0E },
	{ 0x04, 0x05, 0x81, 0x0A, 0x0B, 0x0C, 0x09, 0x0E },
	{ 0x06, 0x07, 0x82, 0x0A, 0x0B, 0x0C, 0x09, 0x0E },
	{ 0xC0, 0xC2, 0x8F, 0xD4, 0xD6, 0xD8, 0xD2, 0xDC },
	{ 0xC4, 0xC6, 0x8B, 0xD4, 0xD6, 0xD8, 0xD2, 0xDC },
	{ 0xC8, 0xCA, 0x89, 0xD4, 0xD6, 0xD8, 0xD2, 0xDC },
	{ 0xCC, 0xCE, 0x8A, 0xD4, 0xD6, 0xD8, 0xD2, 0xDC }
};

dma_t dma[DMA_CHANNELS];

static void *DOSMemAlloc(uint16_t size)
{
	union REGS r;

	r.w.ax = 0x0100;
	r.w.bx = (size + 15) >> 4;
	int386(0x31, &r, &r);

	if (r.x.cflag != 0)
		return NULL;

	bufferSelector = r.w.dx;
	return (void *)(r.w.ax << 4);
}

static void DOSMemFree(int16_t selector)
{
	union REGS r;

	r.w.ax = 0x0101;
	r.w.dx = selector;
	int386(0x31, &r, &r);
}

static void MemLock(void *mem, int16_t size) 
{
	union REGS r;

	r.w.ax = 0x0600;
	r.w.bx = FP_OFF(mem) >> 16;
	r.w.cx = FP_OFF(mem) & 0xFFFF;
	r.w.si = 0;
	r.w.di = size;
	int386(0x31, &r, &r);
}

static void MemUnlock(void *mem, int16_t size) 
{
	union REGS r;

	r.w.ax = 0x0601;
	r.w.bx = FP_OFF(mem) >> 16;
	r.w.cx = FP_OFF(mem) & 0xFFFF;
	r.w.si = 0;
	r.w.di = size;
	int386(0x31, &r, &r);
}

bool DMA_Init(uint8_t channel, uint16_t length, uint8_t mode)
{
	bool bufferAllocated;
	int16_t index, bufferSel[10], i;
	void *buffer;
	dma_t *d;

	if (channel >= DMA_CHANNELS || initialized[channel])
		return false;

	bufferAllocated = false;
	index = 0;

	while (true)
	{
		buffer = (void *)DOSMemAlloc(length);
		if (buffer == NULL)
			break;

		// save every memory allocation performed
		bufferSel[index] = bufferSelector;

		index++;
		if (index == 10)
			break;

		// check page crossing
		if (((FP_OFF(buffer)+length+15) & 0xF0000) == (FP_OFF(buffer) & 0xF0000))
		{
			MemLock(buffer, length);
			bufferAllocated = true;
			break;
		}
	}

	if (!bufferAllocated)
		index++;

	// free all memory blocks crossing a page boundary
	for (i = 0; i < index-1; i++)
		DOSMemFree(bufferSel[i]);

	if (bufferAllocated)
	{
		d = &dma[channel];
		d->buffer = buffer;
		d->mode = mode | (channel & 3);
		d->length = length;

		initialized[channel] = true;
		return true;
	}

	return false;
}

void DMA_Stop(uint8_t channel)
{
	dmaports_t *p;

	if (channel >= DMA_CHANNELS || !active[channel])
		return;

	p = (dmaports_t *)&dmaPorts[channel];

	outp(p->mask, (channel & 3) | 4); // disable DMA channel
	outp(p->flipflop, 0); // reset flip-flop
	
	active[channel] = false;
}

bool DMA_Close(uint8_t channel)
{
	dma_t *d;

	if (channel >= DMA_CHANNELS || !initialized[channel])
		return false;

	d = &dma[channel];

	DMA_Stop(channel);
	MemUnlock(d->buffer, d->length);
	DOSMemFree(bufferSelector);

	initialized[channel] = false;
	return true;
}

bool DMA_Start(uint8_t channel, bool pauseIRQs)
{
	uint16_t page, addr, count;
	dmaports_t *p;
	dma_t *d;

	if (channel >= DMA_CHANNELS || !initialized[channel] || active[channel])
		return false;

	d = &dma[channel];
	p = (dmaports_t *)&dmaPorts[channel];

	page = FP_OFF(d->buffer) >> 16;
	addr = FP_OFF(d->buffer) & 0xFFFF;

	count = d->length;
	if (channel >= 4) // 16-bit DMA
	{
		addr = ((page & 1) << 15) | (addr >> 1);
		count >>= 1;
	}

	count--;

	if (pauseIRQs)
		_disable();

	outp(p->request, 0); // reset DMA request

	outp(p->mask, (channel & 3) | 4); // disable DMA channel

	outp(p->mode, d->mode);
	outp(p->page, page);

	outp(p->flipflop, 0); // reset flip-flop (prepare word access)
	outp(p->addr, (uint8_t)addr);
	outp(p->addr, addr >> 8);
	
	outp(p->flipflop, 0); // reset flip-flop (prepare word access)
	outp(p->count, (uint8_t)count);
	outp(p->count, count >> 8);

	outp(p->clear, 0); // clear mask register

	outp(p->mask, channel & 3); // enable DMA channel

	if (pauseIRQs)
		_enable();

	active[channel] = true;
	return true;
}

void *DMA_GetBuffer(uint8_t channel)
{
	if (channel >= DMA_CHANNELS || !initialized[channel])
		return NULL;

	return dma[channel].buffer;
}
